#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Python inverse kinematics for your robot model based on Pinocchio.
"""

from .configuration import Configuration
from .configuration import assume_configuration
from .configuration import apply_configuration
from .solve_ik import solve_ik

__version__ = "0.2.0"

__all__ = [
    "Configuration",
    "apply_configuration",
    "assume_configuration",
    "models",
    "solve_ik",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Exceptions specific to Pink.
"""


class NotWithinConfigurationLimits(Exception):

    """
    Exception thrown when a robot configuration violates its limits.

    Attributes:
        joint: Index of the joint in the configuration vector.
        value: Invalid value of the joint.
        lower: Minimum allowed value for this joint.
        upper: Maximum allowed value for this joint.
    """

    joint: int
    value: float
    lower: float
    upper: float

    def __init__(
        self,
        joint: int,
        value: float,
        lower: float,
        upper: float,
    ) -> None:
        """
        Create exception.

        Args:
            joint: Index of the joint in the configuration vector.
            value: Invalid value of the joint.
            lower: Minimum allowed value for this joint.
            upper: Maximum allowed value for this joint.
        """
        self.joint = joint
        self.value = value
        self.lower = lower
        self.upper = upper
        self.message = (
            f"Joint {joint} violates configuration limits "
            f"{lower} <= {value} <= {upper}"
        )
        super().__init__(self.message)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Humanoid robot model standing on two feet and reaching with a hand.
"""

import os
import time

import numpy as np
import pinocchio as pin

import pink
import pink.models

from pink.tasks import BodyTask
from pink import solve_ik

import meshcat_shapes


class WavingPose:
    def __init__(self, init: pin.SE3):
        """
        Initialize pose.

        Args:
            init: Initial transform from the wrist frame to the world frame.
        """
        self.init = init

    def at(self, t):
        T = self.init.copy()
        R = T.rotation
        R = np.dot(R, pin.utils.rpyToMatrix(0.0, 0.0, np.pi / 2))
        R = np.dot(R, pin.utils.rpyToMatrix(0.0, np.pi, 0.0))
        T.rotation = R
        T.translation[1] += -0.1 + 0.05 * np.sin(8.0 * t)
        T.translation[2] += 0.5
        return T


if __name__ == "__main__":
    models_dir = os.path.join(os.path.dirname(__file__), "../tests", "models")
    jvrc_description = os.path.join(models_dir, "jvrc_description")
    robot = pink.models.build_from_urdf(jvrc_description)
    viz = pin.visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    robot.setVisualizer(viz, init=False)
    viz.initViewer(open=True)
    viz.loadViewerModel()

    configuration = pink.apply_configuration(robot, robot.q0)
    viz.display(configuration.q)

    left_foot_task = BodyTask(
        "l_ankle", position_cost=1.0, orientation_cost=3.0
    )
    pelvis_task = BodyTask("PELVIS_S", position_cost=1.0, orientation_cost=0.0)
    right_foot_task = BodyTask(
        "r_ankle", position_cost=1.0, orientation_cost=3.0
    )
    right_wrist_task = BodyTask(
        "r_wrist", position_cost=1.0, orientation_cost=3.0
    )
    tasks = [left_foot_task, pelvis_task, right_foot_task, right_wrist_task]

    pelvis_pose = configuration.get_transform_body_to_world("PELVIS_S").copy()
    pelvis_pose.translation[0] += 0.05
    meshcat_shapes.set_frame(viz.viewer["pelvis_pose"])
    viz.viewer["pelvis_pose"].set_transform(pelvis_pose.np)
    pelvis_task.set_target(pelvis_pose)

    left_foot_task.set_target(
        configuration.get_transform_body_to_world("l_ankle")
    )
    right_foot_task.set_target(
        configuration.get_transform_body_to_world("r_ankle")
    )

    right_wrist_pose = WavingPose(
        configuration.get_transform_body_to_world("r_wrist")
    )

    wrist_frame = viz.viewer["right_wrist_pose"]
    meshcat_shapes.set_frame(wrist_frame)

    dt = 5e-3  # [s]
    for t in np.arange(0.0, 10.0, dt):
        right_wrist_task.set_target(right_wrist_pose.at(t))
        wrist_frame.set_transform(right_wrist_pose.at(t).np)
        velocity = solve_ik(configuration, tasks, dt)
        q = configuration.integrate(velocity, dt)
        configuration = pink.apply_configuration(robot, q)
        viz.display(q)
        time.sleep(dt)  # TODO(scaron): proper rate



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of upkie_description.
#
# upkie_description is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# upkie_description is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# upkie_description. If not, see <http://www.gnu.org/licenses/>.

"""
Compute the inertia matrix of a cylinder and print it as URDF.

Source:
    https://en.wikipedia.org/wiki/List_of_moments_of_inertia#List_of_3D_inertia_tensors
"""

import argparse
import sys


def parse_command_line_arguments():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("mass", type=float, help="total mass in [kg]")
    parser.add_argument(
        "radius", type=float, help="radius in [m] in the xy-plane"
    )
    parser.add_argument(
        "length", type=float, help="length in [m] along the z-axis"
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_command_line_arguments()
    mass, radius, length = args.mass, args.radius, args.length
    ixx = mass / 12.0 * (3 * radius ** 2 + length ** 2)
    iyy = mass / 12.0 * (3 * radius ** 2 + length ** 2)
    izz = mass / 2.0 * radius ** 2
    print(f'<cylinder radius="{radius}" length="{length}" />')
    print(f'<mass value="{mass}" />')
    print(f"\n<!-- {' '.join(sys.argv)} -->")
    print(
        f'<inertia ixx="{ixx}" ixy="0" ixz="0" '
        f'iyx="0" iyy="{iyy}" iyz="0" '
        f'izx="0" izy="0" izz="{izz}" />'
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Body task specification.
"""

from typing import Optional, Sequence, Tuple, Union

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from .exceptions import TargetNotSet
from .task import Task
from .utils import body_box_minus


class BodyTask(Task):

    """
    Regulate the pose of a robot body in the world frame.

    Attributes:
        body: Body frame name, typically the link name from the URDF.
        cost: 6D vector that specifies how much each coordinate (in the local
            body frame) contributes to the cost. Position costs come first
            (Pinocchio spatial vector convention) and are in
            :math:`[\\mathrm{cost}] / [\\mathrm{m}]`, where the the unit of
            :math:`[\\mathrm{cost}]` up to the user. They are followed by
            orientation costs in :math:`[\\mathrm{cost}] / [\\mathrm{rad}]`.
            Set a cost to zero to disable the task along a coordinate (no cost
            no effect).
        lm_damping: Unitless scale of the Levenberg-Marquardt (only when
            the error is large) regularization term, which helps when
            targets are unfeasible. Increase this value if the task is too
            jerky under unfeasible targets, but beware that a larger
            damping slows down the task.
        transform_target_to_world: Target pose for the body frame.

    Costs are designed so that position/orientation costs can be compared
    between tasks. For example, if task 1 has a position cost of 1.0 and task 2
    a position cost of 0.1, then a 1 [cm] error in task 1 costs as much as a 10
    [cm] error in task 2.

    Note:
        Dimensionally, the 6D cost vector is a (normalized) force screw and our
        objective function is a (normalized) energy.
    """

    body: str
    cost: np.ndarray
    lm_damping: float
    transform_target_to_world: Optional[pin.SE3]

    def __init__(
        self,
        body: str,
        position_cost: Union[float, Sequence[float]],
        orientation_cost: Union[float, Sequence[float]],
        lm_damping: float = 1e-6,
    ) -> None:
        """
        Define a new body task.

        Args:
            body: Name of the body frame to move to the target pose.
            position_cost: Contribution of position errors to the normalized
                cost, in :math:`[\\mathrm{cost}] / [\\mathrm{m}]`. If this is a
                vector, the cost is anisotropic and each coordinate corresponds
                to an axis in the local body frame.
            orientation_cost: Contribution of orientation errors to the
                normalized cost, in :math:`[\\mathrm{cost}] / [\\mathrm{rad}]`.
                If this is a vector, the cost is anisotropic and each
                coordinate corresponds to an axis in the local body frame.
            lm_damping: Levenberg-Marquardt damping (see class attributes). The
                default value is conservatively low.
        """
        self.body = body
        self.cost = np.ones(6)
        self.lm_damping = lm_damping
        self.transform_target_to_world = None
        #
        self.set_position_cost(position_cost)
        self.set_orientation_cost(orientation_cost)

    def set_position_cost(
        self, position_cost: Union[float, Sequence[float]]
    ) -> None:
        """
        Set a new cost for all 3D position coordinates.

        Args:
            position_cost: Contribution of position errors to the normalized
                cost, in :math:`[\\mathrm{cost}] / [\\mathrm{m}]`. If this is a
                vector, the cost is anisotropic and each coordinate corresponds
                to an axis in the local body frame.
        """
        if isinstance(position_cost, float):
            assert position_cost >= 0.0
        else:  # not isinstance(position_cost, float)
            assert all(cost >= 0.0 for cost in position_cost)
        self.cost[0:3] = position_cost

    def set_orientation_cost(
        self, orientation_cost: Union[float, Sequence[float]]
    ) -> None:
        """
        Set a new cost for all 3D orientation coordinates.

        Args:
            orientation_cost: Contribution of orientation errors to the
                normalized cost, in :math:`[\\mathrm{cost}] / [\\mathrm{rad}]`.
                If this is a vector, the cost is anisotropic and each
                coordinate corresponds to an axis in the local body frame.
        """
        if isinstance(orientation_cost, float):
            assert orientation_cost >= 0.0
        else:  # not isinstance(orientation_cost, float)
            assert all(cost >= 0.0 for cost in orientation_cost)
        self.cost[3:6] = orientation_cost

    def set_target(
        self,
        transform_target_to_world: pin.SE3,
    ) -> None:
        """
        Set task target pose in the world frame.

        Args:
            transform_target_to_world: Transform from the task target frame to
                the world frame.
        """
        self.transform_target_to_world = transform_target_to_world.copy()

    def compute_error_in_body(
        self, configuration: Configuration
    ) -> np.ndarray:
        """
        Compute the body twist error, that is, the (box minus) difference
        between target and current body configuration.

        Args:
            configuration: Robot configuration to read values from.

        Returns:
            Coordinate vector of the body twist error.
        """
        if self.transform_target_to_world is None:
            raise TargetNotSet(f"no target set for body {self.body}")
        transform_body_to_world = configuration.get_transform_body_to_world(
            self.body
        )
        error_in_body: np.ndarray = -body_box_minus(
            self.transform_target_to_world,
            transform_body_to_world,
        )
        return error_in_body

    def compute_task_dynamics(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the matrix :math:`J(q)` and vector :math:`\\alpha e(q)` such
        that the task dynamics are:

        .. math::

            J(q) \\Delta q = \\alpha e(q)

        The Jacobian matrix is :math:`J(q) \\in \\mathbb{R}^{6 \\times n}`,
        with :math:`n` the dimension of the robot's tangent space, and the
        error vector is :math:`e(q) \\in \\mathbb{R}^6`.

        See :func:`Task.compute_task_dynamics` for more context.

        Args:
            configuration: Robot configuration to read values from.

        Returns:
            Pair :math:`(J, \\alpha e)` of Jacobian matrix and error vector,
            both expressed in the body frame.
        """
        jacobian_in_body = configuration.get_body_jacobian(self.body)
        error_in_body = self.compute_error_in_body(configuration)
        return jacobian_in_body, self.gain * error_in_body

    def compute_qp_objective(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the Hessian matrix :math:`H` and linear vector :math:`c` such
        that the contribution of the task to the QP objective is:

        .. math::

            \\| J \\Delta q - \\alpha e \\|_{W}^2
            = \\frac{1}{2} \\Delta q^T H \\Delta q + c^T q

        The weight matrix :math:`W \\in \\mathbb{R}^{6 \\times 6}` combines
        position and orientation costs. The unit of the overall contribution is
        :math:`[\\mathrm{cost}]^2`. The configuration displacement
        :math:`\\Delta q` is the output of inverse kinematics (we divide it by
        :math:`\\Delta t` to get a commanded velocity).

        Args:
            configuration: Robot configuration to read values from.

        Returns:
            Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
            objective.

        See also:
            Levenberg-Marquardt damping is described in
            "Solvability-Unconcerned Inverse Kinematics by the
            Levenberg-Marquardt Method" (Sugihara, 2011). The dimensional
            analysis in this class is our own.
        """
        jacobian, error = self.compute_task_dynamics(configuration)
        weight = np.diag(self.cost)  # [cost] * [twist]^{-1}
        weighted_jacobian = weight @ jacobian  # [cost]
        weighted_error = weight @ error  # [cost]
        mu = self.lm_damping * weighted_error @ weighted_error  # [cost]^2
        eye_tg = configuration.tangent.eye
        # Our Levenberg-Marquardt damping `mu * eye_tg` is isotropic in the
        # robot's tangent space. If it helps we can add a tangent-space scaling
        # to damp the floating base differently from joint angular velocities.
        H = weighted_jacobian.T @ weighted_jacobian + mu * eye_tg
        c = -weighted_error.T @ weighted_jacobian
        return (H, c)

    def __repr__(self):
        """
        Human-readable representation of the task.
        """
        return (
            f"BodyTask({self.body}, "
            f"gain={self.gain}, "
            f"orientation_cost={self.cost[3:6]}, "
            f"position_cost={self.cost[0:3]}, "
            f"target={self.transform_target_to_world})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Function to solve inverse kinematics.
"""

from typing import Iterable, Tuple

import numpy as np

from qpsolvers import solve_qp

from .configuration import Configuration
from .limits import compute_velocity_limits
from .tasks import Task


def compute_qp_objective(
    configuration: Configuration, tasks: Iterable[Task], damping: float
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the Hessian matrix :math:`H` and linear vector :math:`c` of the
    QP objective function:

    .. math::

        \\frac{1}{2} \\Delta q^T H \\Delta q + c^T q

    The configuration displacement :math:`\\Delta q` is the output of inverse
    kinematics (we divide it by :math:`\\Delta t` to get a commanded velocity).

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks to fulfill at (weighted) best.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            `[cost]^2 / [tangent]` where `[tangent]` is "the" unit of robot
            velocities. Improves numerical stability, but larger values slow
            down all tasks.

    Returns:
        Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
        objective.
    """
    H = damping * configuration.tangent.eye
    c = configuration.tangent.zeros
    for task in tasks:
        H_task, c_task = task.compute_qp_objective(configuration)
        H += H_task
        c += c_task
    return (H, c)


def solve_ik(
    configuration: Configuration,
    tasks: Iterable[Task],
    dt: float,
    damping: float = 1e-12,
    solver: str = "quadprog",
) -> np.ndarray:
    """
    Compute a velocity tangent to the current robot configuration that
    satisfies at (weighted) best a given set of kinematic tasks.

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\\mathrm{cost}]^2 / [\\mathrm{tangent}]` where
            :math:`[\\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks.
        solver: Backend quadratic programming solver.

    Returns:
        Velocity :math:`v` in tangent space.

    Raises:
        NotWithinConfigurationLimits: if the current configuration is not
            within limits.

    Note:
        Our Tikhonov damping is isotropic despite tangent velocities not being
        homogeneous. If it helps we can add a tangent-space scaling to damp the
        floating base differently from joint angular velocities.
    """
    configuration.check_limits()
    H, c = compute_qp_objective(configuration, tasks, damping)
    v_max, v_min = compute_velocity_limits(configuration, dt)
    tangent_eye = configuration.tangent.eye
    A = np.vstack([tangent_eye, -tangent_eye])
    b = np.hstack([v_max * dt, -v_min * dt])
    Delta_q = solve_qp(H, c, A, b, solver=solver)
    assert Delta_q is not None
    v: np.ndarray = Delta_q / dt
    return v



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
TODO(scaron): move these utility functions somewhere logical.
"""

import meshcat
import numpy as np


def attach_axes(
    frame: meshcat.Visualizer,
    length: float = 0.05,
    thickness: float = 0.002,
    opacity: float = 1.0,
) -> None:
    """
    Attach a set of three basis axes to a MeshCat handle.

    Args:
        frame: MeshCat handle to attach the basis axes to.
        length: Length of axis unit vectors.
        thickness: Thickness of axis unit vectors.
        opacity: Opacity of all three unit vectors.

    Note:
        As per the de-facto standard (Blender, OpenRAVE, RViz, ...), the
        x-axis is red, the y-axis is green and the z-axis is blue.
    """
    direction_names = ["x", "y", "z"]
    colors = [0xFF0000, 0x00FF00, 0x0000FF]
    rotation_axes = [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
    position_cylinder_in_body = 0.5 * length * np.eye(3)
    for i in range(3):
        dir_name = direction_names[i]
        material = meshcat.geometry.MeshLambertMaterial(
            color=colors[i], opacity=opacity
        )
        transform_cylinder_to_body = meshcat.transformations.rotation_matrix(
            np.pi / 2, rotation_axes[i]
        )
        transform_cylinder_to_body[0:3, 3] = position_cylinder_in_body[i]
        cylinder = meshcat.geometry.Cylinder(length, thickness)
        frame[dir_name].set_object(cylinder, material)
        frame[dir_name].set_transform(transform_cylinder_to_body)


def set_frame(
    frame: meshcat.Visualizer,
    axis_length: float = 0.1,
    axis_thickness: float = 0.005,
    opacity: float = 1.0,
    origin_color: int = 0x000000,
    origin_radius: float = 0.01,
) -> None:
    """
    Set MeshCat handle to a frame, represented by an origin and three axes.

    Args:
        frame: MeshCat handle to attach the frame to.
        axis_length: Length of axis unit vectors, in [m].
        axis_thickness: Thickness of axis unit vectors, in [m].
        opacity: Opacity of all three unit vectors.
        origin_color: Color of the origin sphere.
        origin_radius: Radius of the frame origin sphere, in [m].

    Note:
        As per the de-facto standard (Blender, OpenRAVE, RViz, ...), the
        x-axis is red, the y-axis is green and the z-axis is blue.
    """
    material = meshcat.geometry.MeshLambertMaterial(
        color=origin_color, opacity=opacity
    )
    sphere = meshcat.geometry.Sphere(origin_radius)
    frame.set_object(sphere, material)
    attach_axes(frame, axis_length, axis_thickness, opacity)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test inverse kinematics.
"""

import os
import unittest

import numpy as np
import pinocchio as pin

from numpy.linalg import norm

import pink

from pink import apply_configuration
from pink import solve_ik
from pink.exceptions import NotWithinConfigurationLimits
from pink.models import build_from_urdf
from pink.tasks import BodyTask


class TestSolveIK(unittest.TestCase):

    """
    Test fixture for the solve_ik function.
    """

    def setUp(self):
        models_dir = os.path.join(os.path.dirname(__file__), "models")
        jvrc_description = os.path.join(models_dir, "jvrc_description")
        upkie_description = os.path.join(models_dir, "upkie_description")
        self.jvrc_description = jvrc_description
        self.upkie_description = upkie_description

    def test_no_task(self):
        """
        Raise an error when the robot body is not found.
        """
        robot = build_from_urdf(self.jvrc_description)
        configuration = apply_configuration(robot, robot.q0)
        tasks = []
        v = solve_ik(configuration, tasks, dt=1e-3)
        self.assertTrue(np.allclose(v, np.zeros(robot.nv)))

    def test_single_task_fulfilled(self):
        """
        Velocity is zero when the only task is already fulfilled.
        """
        robot = build_from_urdf(self.upkie_description)
        configuration = apply_configuration(robot, robot.q0)
        task = BodyTask(
            "left_contact", position_cost=1.0, orientation_cost=1.0
        )
        task.set_target(
            configuration.get_transform_body_to_world("left_contact")
        )
        velocity = solve_ik(configuration, [task], dt=5e-3)
        self.assertTrue(np.allclose(velocity, 0.0))

    def test_single_task_convergence(self):
        """
        Integrating IK velocities makes a single task converge to its target.
        """
        robot = build_from_urdf(self.upkie_description)
        configuration = apply_configuration(robot, robot.q0)
        task = BodyTask(
            "left_contact", position_cost=1.0, orientation_cost=1.0
        )
        transform_init_to_world = configuration.get_transform_body_to_world(
            "left_contact"
        )
        transform_target_to_init = pin.SE3(
            np.eye(3), np.array([0.0, 0.0, 0.1])
        )
        transform_target_to_world = (
            transform_init_to_world * transform_target_to_init
        )
        task.set_target(transform_target_to_world)
        dt = 5e-3  # [s]
        velocity = solve_ik(configuration, [task], dt)

        # Initially we are nowhere near the target and moving
        self.assertFalse(np.allclose(velocity, 0.0))
        self.assertAlmostEqual(
            norm(task.compute_error_in_body(configuration)),
            0.1,
        )
        self.assertFalse(
            configuration.get_transform_body_to_world("left_contact").isApprox(
                transform_target_to_world, prec=1e-4
            )
        )

        last_error = 1e6
        for nb_steps in range(42):
            error = norm(task.compute_error_in_body(configuration))
            if error < 1e-6 and np.allclose(velocity, 0.0):
                break
            self.assertLess(error, last_error)  # error stictly decreases
            last_error = error
            q = configuration.integrate(velocity, dt)
            configuration = apply_configuration(robot, q)
            velocity = solve_ik(configuration, [task], dt)

        # After nb_steps we are at the target and not moving
        self.assertTrue(np.allclose(velocity, 0.0))
        self.assertAlmostEqual(
            norm(task.compute_error_in_body(configuration)),
            0.0,
        )
        self.assertTrue(
            configuration.get_transform_body_to_world("left_contact").isApprox(
                transform_target_to_world, prec=1e-8
            )
        )
        self.assertLess(nb_steps, 3)

    def test_three_tasks_fulfilled(self):
        """
        No motion when all targets are reached.
        """
        robot = build_from_urdf(self.jvrc_description)
        configuration = apply_configuration(robot, robot.q0)
        left_ankle_task = BodyTask(
            "l_ankle", position_cost=1.0, orientation_cost=3.0
        )
        right_ankle_task = BodyTask(
            "r_ankle", position_cost=1.0, orientation_cost=3.0
        )
        pelvis_task = BodyTask(
            "PELVIS_S", position_cost=1.0, orientation_cost=3.0
        )

        left_ankle_task.set_target(
            configuration.get_transform_body_to_world("l_ankle")
        )
        right_ankle_task.set_target(
            configuration.get_transform_body_to_world("r_ankle")
        )
        pelvis_task.set_target(
            configuration.get_transform_body_to_world("PELVIS_S")
        )

        tasks = [pelvis_task, left_ankle_task, right_ankle_task]
        velocity = solve_ik(configuration, tasks, dt=5e-3)
        self.assertTrue(np.allclose(velocity, 0.0))

    def test_checks_configuration_limits(self):
        """
        IK checks for configuration limits.
        """
        robot = build_from_urdf(self.upkie_description)
        q = robot.q0
        q[7] = 20  # above limit for Upkie's first joint
        configuration = pink.apply_configuration(robot, q)
        with self.assertRaises(NotWithinConfigurationLimits):
            solve_ik(configuration, [], dt=1.0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Posture task specification.
"""

from typing import Optional, Tuple

import numpy as np

from ..configuration import Configuration
from .exceptions import TargetNotSet
from .task import Task


class PostureTask(Task):

    """
    Regulate joint angles to a desired posture, *i.e.* a vector of actuated
    joint angles. Floating base coordinates are not affected by this task.

    Attributes:
        cost: joint angular error cost in
            :math:`[\\mathrm{cost}] / [\\mathrm{rad}]`.
        target_q: Target vector in the configuration space.

    A posture task is typically used for regularization as it has a steady
    rank. For instance, when Upkie's legs are stretched and the Jacobian of its
    contact frames become singular, the posture task will drive the knees
    toward a preferred orientation.
    """

    cost: float
    target_q: Optional[np.ndarray]

    def __init__(self, cost: float) -> None:
        """
        Create task.

        Args:
            cost: joint angular error cost in
                :math:`[\\mathrm{cost}] / [\\mathrm{rad}]`.

        Note:
            We assume that the first seven coordinates of the configuration are
            for the floating base.
        """
        self.cost = cost
        self.target_q = None

    def set_target(self, target_q: np.ndarray) -> None:
        """
        Set task target pose in the world frame.

        Args:
            target_q: Target vector in the configuration space.
        """
        self.target_q = target_q.copy()

    def compute_task_dynamics(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the matrix :math:`J(q)` and vector :math:`\\alpha e(q)` such
        that the task dynamics are:

        .. math::

            J(q) \\Delta q = \\alpha e(q)

        The Jacobian matrix is :math:`J(q) \\in \\mathbb{R}^{n \\times n}`,
        with :math:`n` the dimension of the robot's tangent space, and the
        error vector is :math:`e(q) \\in \\mathbb{R}^n`. Both depend on the
        configuration :math:`q` of the robot.

        See :func:`Task.compute_task_dynamics` for more context.

        Args:
            configuration: Robot configuration to read kinematics from.

        Returns:
            Pair :math:`(J, \\alpha e)` of Jacobian matrix and error vector,
            both expressed in the body frame.
        """
        if self.target_q is None:
            raise TargetNotSet("no posture target")
        # TODO(scaron): handle models without floating base joint
        jacobian = configuration.tangent.eye[6:, :]
        error = (self.target_q - configuration.q)[7:]
        return (jacobian, self.gain * error)

    def compute_qp_objective(
        self, configuration: Configuration
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the Hessian matrix :math:`H` and linear vector :math:`c` such
        that the contribution of the task to the QP objective is:

        .. math::

            \\| J \\Delta q - \\alpha e \\|_{W}^2
            = \\frac{1}{2} \\Delta q^T H \\Delta q + c^T q

        The weight matrix :math:`W \\in \\mathbb{R}^{n \\times n}` weighs and
        normalizes task coordinates to the same unit. The unit of the overall
        contribution is :math:`[\\mathrm{cost}]^2`. The configuration
        displacement :math:`\\Delta q` is the output of inverse kinematics (we
        divide it by :math:`\\Delta t` to get a commanded velocity).

        Args:
            robot: Robot model and configuration.

        Returns:
            Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
            objective.
        """
        jacobian, error = self.compute_task_dynamics(configuration)
        weighted_jacobian = self.cost * jacobian  # [cost]
        weighted_error = self.cost * error  # [cost]
        H = weighted_jacobian.T @ weighted_jacobian
        c = -weighted_error.T @ weighted_jacobian
        return (H, c)

    def __repr__(self):
        """
        Human-readable representation of the task.
        """
        return f"PostureTask(cost={self.cost}, gain={self.gain})"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joint limits implemented as inequality constraints.
"""

from typing import Tuple

import numpy as np

from .configuration import Configuration


def compute_velocity_limits(
    configuration: Configuration,
    dt: float,
    config_limit_gain: float = 0.5,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the configuration-dependent velocity limits:

    .. math::

        v_{min}(q) \\leq v \\leq v_{max}(q)

    where :math:`q \\in {\\cal C}` is the robot's configuration and :math:`v
    \\in T_q({\\cal C})` is the velocity in the tangent space at :math:`q`.
    These velocity limits combine:

    - Configuration-independent joint velocity limits
      :math:`|\\dot{q}| \\leq \\dot{q}_{lim}`.
    - Time-derivatives of the configuration limit
      :math:`q_{min} \\leq q \\leq q_{max}`.

    Args:
        configuration: Robot configuration to read kinematics from.
        dt: Integration timestep in [s].
        config_limit_gain: gain between 0 and 1 to steer away from
            configuration limits. It is described in "Real-time prioritized
            kinematic control under inequality constraints for redundant
            manipulators" (Kanoun, 2012). More details in [this
            writeup](https://scaron.info/teaching/inverse-kinematics.html).

    Returns:
        Pair $(v_{max}(q), v_{min}(q))$ of velocity lower and upper bounds.
    """
    v_max = +1000.0 * configuration.tangent.ones
    v_min = -1000.0 * configuration.tangent.ones
    return v_max, v_min



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pink adds a notion of *configured* robot to indicate that forward kinematics
functions have been run, so that frame transforms and frame Jacobians used for
IK can be queried from the robot's data.
"""

import numpy as np
import pinocchio as pin

from .exceptions import NotWithinConfigurationLimits


class Tangent:

    """
    Represent the tangent space at a given robot configuration.
    """

    __model: pin.Model

    def __init__(self, model: pin.Model):
        self.__model = model

    @property
    def eye(self):
        """
        Identity matrix of the tangent space.
        """
        return np.eye(self.__model.nv)

    @property
    def ones(self):
        """
        Vector of ones in the tangent space.
        """
        return np.ones(self.__model.nv)

    @property
    def zeros(self):
        """
        Vector of zeros in the tangent space.
        """
        return np.zeros(self.__model.nv)


class Configuration:

    """
    Type annotation indicating that quantities that depend on the configuration
    :math:`q` have been computed.

    In Pink, this type enables access to frame transforms and frame Jacobians.
    We rely on typing to make sure the proper forward kinematics functions have
    been called beforehand. In Pinocchio, these functions are:

    - ``pin.computeJointJacobians(model, data, configuration)``
    - ``pin.updateFramePlacements(model, data)``

    The former computes the full model Jacobian into ``data.J``. (It also
    computes forward kinematics, so there is no need to further call
    ``pin.forwardKinematics(model, data, configuration)``.) The latter updates
    frame placements.

    Notes:
        This class is meant to be used as a subclass of pin.RobotWrapper, not
        wrap it. However, right now pin.RobotWrapper does not have a shallow
        copy constructor. TODO(scaron): bring it up upstream.

    Attributes:
        data: Data with kinematics matching the configuration vector
            :data:`Configuration.q`.
        model: Kinodynamic model.
        q: Configuration vector matching the kinematics in
            :data:`Configuration.data`.
    """

    data: pin.Data
    model: pin.Model
    q: np.ndarray

    def __init__(self, model: pin.Model, data: pin.Data, q: np.ndarray):
        self.data = data
        self.model = model
        self.q = q
        self.tangent = Tangent(model)

    def check_limits(self) -> None:
        """
        Check that the current configuration is within limits.

        Raises:
            NotWithinConfigurationLimits: if the current configuration is
                within limits.
        """
        q_max = self.model.upperPositionLimit
        q_min = self.model.lowerPositionLimit
        for i in range(7, self.model.nq):
            if q_max[i] <= q_min[i] + 1e-6:  # no limit
                continue
            if self.q[i] < q_min[i] or self.q[i] > q_max[i]:
                raise NotWithinConfigurationLimits(
                    i,
                    self.q[i],
                    q_min[i],
                    q_max[i],
                )

    def get_body_jacobian(self, body: str) -> np.ndarray:
        """
        Compute the Jacobian matrix :math:`{}_B J_{WB}` of the body velocity
        :math:`{}_B v_{WB}`:

        .. math::

            {}_B v_{WB} = {}_B J_{WB} \\dot{q}

        Args:
            body: Body frame name, typically the link name from the URDF.

        Returns:
            Jacobian :math:`{}_B J_{WB}` of the body twist.

        When the robot model includes a floating base
        (pin.JointModelFreeFlyer), the configuration vector :math:`q` consists
        of:

        - ``q[0:3]``: position in [m] of the floating base in the inertial
          frame, formatted as :math:`[p_x, p_y, p_z]`.
        - ``q[3:7]``: unit quaternion for the orientation of the floating base
          in the inertial frame, formatted as :math:`[q_x, q_y, q_z, q_w]`.
        - ``q[7:]``: joint angles in [rad].
        """
        body_id = self.model.getBodyId(body)
        J: np.ndarray = pin.getFrameJacobian(
            self.model, self.data, body_id, pin.ReferenceFrame.LOCAL
        )
        return J

    def get_transform_body_to_world(self, body: str) -> pin.SE3:
        """
        Get the pose of a body frame of the robot in its current configuration.

        Args:
            body: Body frame name, typically the link name from the URDF.

        Returns:
            Current transform from body frame to world frame.

        Raises:
            KeyError: if the body name is not found in the robot model.
        """
        body_id = self.model.getBodyId(body)
        try:
            return self.data.oMf[body_id].copy()
        except IndexError as index_error:
            raise KeyError(
                f'Body "{body}" not found in robot model'
            ) from index_error

    def integrate(self, velocity, dt) -> np.ndarray:
        """
        Integrate a velocity starting from the current configuration.

        Args:
            velocity: Velocity in tangent space.
            dt: Integration duration in [s].

        Returns:
            New configuration after integration.
        """
        return pin.integrate(self.model, self.q, velocity * dt)


def assume_configuration(
    robot: pin.RobotWrapper, q: np.ndarray
) -> Configuration:
    """
    Assume that the provided robot wrapper has already been configured.

    Args:
        robot: Robot wrapper with its initial data.
        q: Configuration matching the robot wrapper's data.

    Returns:
        Robot configuration.
    """
    return Configuration(robot.model, robot.data, q)


def apply_configuration(
    robot: pin.RobotWrapper, q: np.ndarray
) -> Configuration:
    """
    Run forward kinematics on a robot wrapper.

    Args:
        robot: Robot wrapper with its initial data.
        q: Configuration vector to apply.

    Returns:
        Configured robot.
    """
    pin.computeJointJacobians(robot.model, robot.data, q)
    pin.updateFramePlacements(robot.model, robot.data)
    return Configuration(robot.model, robot.data, q)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test fixture for other library features.
"""

import unittest

from pink.models import UnknownModel
from pink.models import build_from_urdf


class TestConfiguration(unittest.TestCase):
    def test_build_from_unknown_model(self):
        """
        Raise exception when model is not found.
        """
        with self.assertRaises(UnknownModel):
            build_from_urdf("foo_smash_super_bar")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test base class for kinematic tasks.
"""

import os
import unittest

from pink.models import build_from_urdf
from pink.tasks import Task


class TestTask(unittest.TestCase):
    def setUp(self):
        """
        Prepare test fixture.
        """
        models_dir = os.path.join(os.path.dirname(__file__), "models")
        jvrc_description = os.path.join(models_dir, "jvrc_description")
        self.jvrc_description = jvrc_description

    def test_task_repr(self):
        """
        String representation reports the task gain.
        """
        task = Task()
        self.assertTrue("gain=" in repr(task))

    def test_task_not_implemented(self):
        """
        Raise an error when the robot body is not found.
        """
        robot = build_from_urdf(self.jvrc_description)
        task = Task()
        self.assertIsNotNone(task)
        with self.assertRaises(NotImplementedError):
            task.compute_task_dynamics(robot)
        with self.assertRaises(NotImplementedError):
            task.compute_qp_objective(robot)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the Configuration type.
"""

import os
import unittest

import numpy as np

import pink

from pink.exceptions import NotWithinConfigurationLimits
from pink.models import build_from_urdf


class TestConfiguration(unittest.TestCase):
    def setUp(self):
        """
        Prepare test fixture.
        """
        models_dir = os.path.join(os.path.dirname(__file__), "models")
        jvrc_description = os.path.join(models_dir, "jvrc_description")
        self.jvrc_description = jvrc_description

    def test_assume_configuration(self):
        """
        Assuming a configuration does not change data.
        """
        robot = build_from_urdf(self.jvrc_description)
        robot.data.J.fill(42.0)
        configuration = pink.assume_configuration(robot, robot.q0)
        self.assertTrue(np.allclose(configuration.data.J, 42.0))

    def test_apply_configuration(self):
        """
        Applying a configuration computes Jacobians.
        """
        robot = build_from_urdf(self.jvrc_description)
        robot.data.J.fill(42.0)
        configuration = pink.apply_configuration(robot, robot.q0)
        J_check = np.array(
            [
                [
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.096,
                    0.389,
                    0.0,
                    0.746,
                    -0.0,
                    -0.0,
                    -0.096,
                    0.389,
                    0.0,
                    0.746,
                    0.0,
                    -0.192,
                    0.0,
                    -0.522,
                    0.0,
                    0.24,
                    -0.217,
                    0.24,
                    0.0,
                    0.24,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.645,
                    -0.522,
                    -0.0,
                    -0.24,
                    -0.217,
                    -0.24,
                    0.0,
                    -0.24,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.746,
                    -0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.746,
                    -0.0,
                    0.0,
                    0.0,
                    0.192,
                    0.0,
                    0.522,
                    0.0,
                    0.0,
                    -0.0,
                    -0.022,
                    -0.0,
                    -0.124,
                    -0.169,
                    -0.124,
                    -0.169,
                    -0.106,
                    -0.151,
                    0.003,
                    0.645,
                    0.0,
                    0.0,
                    0.522,
                    0.0,
                    0.0,
                    -0.0,
                    -0.022,
                    -0.0,
                    -0.124,
                    -0.169,
                    -0.124,
                    -0.169,
                    -0.106,
                    -0.151,
                ],
                [
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.096,
                    0.0,
                    -0.02,
                    -0.096,
                    0.02,
                    0.0,
                    0.096,
                    0.0,
                    -0.02,
                    0.096,
                    0.02,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    -0.24,
                    0.0,
                    0.004,
                    0.0,
                    -0.24,
                    0.0,
                    -0.246,
                    -0.246,
                    -0.246,
                    -0.246,
                    -0.217,
                    -0.217,
                    -0.0,
                    -0.0,
                    -0.003,
                    0.0,
                    0.24,
                    0.0,
                    0.004,
                    0.0,
                    0.24,
                    0.0,
                    0.246,
                    0.246,
                    0.246,
                    0.246,
                    0.217,
                    0.217,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
            ]
        )
        self.assertTrue(np.allclose(configuration.data.J, J_check))

    def test_transform_found(self):
        """
        Return the pose of an existing robot body.
        """
        robot = build_from_urdf(self.jvrc_description)
        configuration = pink.apply_configuration(robot, robot.q0)
        transform_pelvis_to_world = configuration.get_transform_body_to_world(
            "PELVIS_S"
        )
        self.assertTrue(
            np.allclose(
                transform_pelvis_to_world.np[3, :],
                np.array([0.0, 0.0, 0.0, 1.0]),
            )
        )

    def test_transform_not_found(self):
        """
        Raise an error when the request robot body is not found.
        """
        robot = build_from_urdf(self.jvrc_description)
        configuration = pink.apply_configuration(robot, robot.q0)
        with self.assertRaises(KeyError):
            configuration.get_transform_body_to_world("foo")

    def test_check_limits(self):
        """
        Raise an error if and only if a joint limit is exceened.
        """
        robot = build_from_urdf(self.jvrc_description)
        q = robot.q0
        configuration = pink.apply_configuration(robot, q)
        configuration.check_limits()
        q[-10] += 1e4  # TODO(scaron): this is actually undesirable!
        with self.assertRaises(NotWithinConfigurationLimits):
            configuration.check_limits()

    def test_tangent_eye(self):
        """
        Configuration's tangent eye is an identity matrix.
        """
        robot = build_from_urdf(self.jvrc_description)
        configuration = pink.apply_configuration(robot, robot.q0)
        v = np.array([i for i in range(robot.model.nv)])
        self.assertTrue(np.allclose(configuration.tangent.eye.dot(v), v))

    def test_tangent_ones(self):
        """
        Configuration's tangent ones is a vector of 1.0's.
        """
        robot = build_from_urdf(self.jvrc_description)
        configuration = pink.apply_configuration(robot, robot.q0)
        self.assertEqual(np.sum(configuration.tangent.ones), robot.model.nv)

    def test_tangent_zeros(self):
        """
        Configuration's tangent ones is a vector of 0.0's.
        """
        robot = build_from_urdf(self.jvrc_description)
        configuration = pink.apply_configuration(robot, robot.q0)
        self.assertAlmostEqual(np.sum(configuration.tangent.zeros), 0.0)
        self.assertEqual(len(configuration.tangent.zeros), robot.model.nv)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
`Upkie`_ is a homemade wheeled biped robot that can balance, crouch and turn
around. It proudly stands on broomsticks and 3D printed parts (^_^)

.. _Upkie: https://www.youtube.com/watch?v=8b36XcCgh7s
"""

import os

import pinocchio as pin


def build_upkie_from_urdf(path: str) -> pin.RobotWrapper:
    """
    Build the Upkie model from its URDF.

    Args:
        path: Path to an `upkie_description`_ folder.

    .. _upkie_description: https://github.com/tasts-robots/upkie_description
    """
    abspath = os.path.abspath(path)
    return pin.RobotWrapper.BuildFromURDF(
        filename=os.path.join(abspath, "urdf", "upkie.urdf"),
        package_dirs=[os.path.dirname(abspath)],
        root_joint=pin.JointModelFreeFlyer(),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test fixture for the posture task.
"""

import os
import unittest

import numpy as np

from pink.tasks import PostureTask
from pink.tasks import TargetNotSet

from .mock_configuration import MockConfiguration


class TestPostureTask(unittest.TestCase):

    """
    Test consistency of the posture task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """
        Prepare test fixture.
        """
        models_dir = os.path.join(os.path.dirname(__file__), "models")
        jvrc_description = os.path.join(models_dir, "jvrc_description")
        self.jvrc_description = jvrc_description
        self.mock_configuration = MockConfiguration()

    def test_target_not_set(self):
        """
        Raise an exception when the target is not set.
        """
        task = PostureTask(cost=1.0)
        with self.assertRaises(TargetNotSet):
            task.compute_task_dynamics(self.mock_configuration)

    def test_target_set_properly(self):
        """
        Return target properly once it's set.
        """
        task = PostureTask(cost=1.0)
        task.set_target(self.mock_configuration.q)
        self.assertIsNotNone(task.target_q)
        if task.target_q is None:  # help mypy
            return
        self.assertTrue(np.allclose(self.mock_configuration.q, task.target_q))

    def test_target_is_a_copy(self):
        """
        Target is saved as a copy, not a reference to the original.
        """
        task = PostureTask(cost=1.0)
        q = self.mock_configuration.q
        task.set_target(q)
        y = q[2]
        q[2] += 12.0
        if task.target_q is None:  # help mypy
            return
        self.assertAlmostEqual(task.target_q[2], y)
        self.assertNotAlmostEqual(task.target_q[2], q[2])

    def test_zero_error_when_target_at_body(self):
        """
        Error is zero when the target and body are at the same location.
        """
        task = PostureTask(cost=1.0)
        q = self.mock_configuration.q
        task.set_target(q)  # error == 0
        _, e = task.compute_task_dynamics(self.mock_configuration)
        self.assertLess(np.linalg.norm(e), 1e-10)

    def test_unit_cost_qp_objective(self):
        """
        A unit cost vector means the QP objective is exactly :math:`(H, c) =
        (J^T J, -e^T J)`.
        """
        task = PostureTask(cost=1.0)
        q = self.mock_configuration.q
        task.set_target(q)
        self.mock_configuration.q[1] += 1.0
        self.mock_configuration.q[3] += 1.0
        self.mock_configuration.q[5] += 1.0
        J, e = task.compute_task_dynamics(self.mock_configuration)
        H, c = task.compute_qp_objective(self.mock_configuration)
        self.assertTrue(np.allclose(J.T @ J, H))
        self.assertTrue(np.allclose(-e.T @ J, c))

    def test_zero_cost_same_as_disabling_task(self):
        """
        The task has no effect when its cost is zero.
        """
        task = PostureTask(cost=0.0)
        q = self.mock_configuration.q
        task.set_target(q)
        J, e = task.compute_task_dynamics(self.mock_configuration)
        H, c = task.compute_qp_objective(self.mock_configuration)
        qd = np.random.random(J.shape[1:])
        cost = qd.T @ H @ qd + c @ qd
        self.assertAlmostEqual(cost, 0.0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lie algebra utility functions.
"""

import numpy as np
import pinocchio as pin


def box_minus(plus_transform: pin.SE3, minus_transform: pin.SE3) -> np.ndarray:
    """
    Compute the box minus between two transforms:

    .. math::

        T_1 \\boxminus T_2 = \\log(T_1 \\cdot T_2^{-1})

    This operator allows us to think about orientation "differences" as
    similarly as possible to position differences, but mind the frames! Its
    formula has two use cases, depending on whether the common frame :math:`C`
    between the two transforms is their source or their target.

    When the common frame is the target, denoting by :math:`T_{CP}` the
    transform from frame :math:`P` (source) to frame :math:`C` (target), the
    resulting twist is expressed in the target frame:

    .. math::

        {}_C \\xi_{CM} = T_{CP} \\boxminus T_{CM}

    When the common frame is the source frame, denoting by :math:`T_{MC}` the
    transform from frame :math:`C` (source) to frame :math:`M` (target), the
    resulting twist is expressed in the target frame of the transform on the
    right-hand side of the operator:

    .. math::

        -{}_M \\xi_{M} = T_{PC} \\boxminus T_{MC}

    Args:
        plus_transform: Transform :math:`T_1` on the left-hand side of the box
            minus operator.
        minus_transform: Transform :math:`T_2` on the right-hand side of the
            box minus operator.

    Returns:
        In the first case :math:`T_{CP} \\boxminus T_{CM}`, the outcome is a
        spatial twist :math:`{}_C \\xi_{CM}` expressed in the common frame
        :math:`C`.

        In the second case :math:`T_{PC} \\boxminus T_{MC}`, the outcome is a
        body twist :math:`-{}_M \\xi_{CM}` (mind the unitary minus).

    Note:
        Prefer using :func:`pink.tasks.utils.body_box_minus` to calling this
        function in the second use case :math:`T_{PC} \\boxminus T_{MC}`.
    """
    diff_array = plus_transform.act(minus_transform.inverse())
    twist: np.ndarray = pin.log(diff_array).vector
    return twist


def body_box_minus(
    transform_plus_to_common: pin.SE3,
    transform_minus_to_common: pin.SE3,
) -> np.ndarray:
    """
    Compute the box minus between two poses :math:`P` (plus) and :math:`M`
    (minus) represented by their transforms :math:`T_{CP}` and :math:`T_{CM}`
    to a common frame :math:`C`.

    .. math::

        {}_M \\xi_{CM} = -(T_{PC} \\boxminus T_{MC})
                       = -(T_{CP}^{-1} \\boxminus T_{CM}^{-1})

    Args:
        transform_plus_to_common: Transform :math:`T_{CP}` from the "plus"
            frame to the common frame.
        transform_minus_to_common: Transform :math:`T_{CM}` from the "minus"
            frame to the common frame.

    Returns:
        Body twist :math:`{}_M \\xi_{CM}`. If :math:`C` is inertial (and has
        zero velocity) we can simply write it :math:`{}_M \\xi_M` like other
        `body screws
        <https://scaron.info/robotics/screw-theory.html#body-screws>`_.
    """
    return box_minus(
        transform_plus_to_common.inverse(),
        transform_minus_to_common.inverse(),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mock configuration used by test fixtures.
"""

from typing import Union

import numpy as np
import pinocchio as pin

import pink


class MockModel:

    """
    Mock robot model, 6 joints and a floating base.
    """

    nq: int = 13
    nv: int = 12


class MockConfiguration(pink.Configuration):

    """
    Mock robot configuration.
    """

    def __init__(self):
        super(MockConfiguration, self).__init__(MockModel(), None, None)
        self.transforms = {}
        self.q = np.ones(self.model.nq)

    def get_transform_body_to_world(self, body: str) -> pin.SE3:
        """
        Generate random poses with memoization.

        Args:
            body: Each name will return the same pose.

        Returns:
            Body pose in the world frame.
        """
        if body not in self.transforms:
            self.transforms[body] = pin.SE3.Random()
        return self.transforms[body]

    def move_body(self, body: str, translate: Union[list, np.ndarray]) -> None:
        self.transforms[body].translation += np.array(translate)

    def get_body_jacobian(self, body: str) -> np.ndarray:
        return np.ones((6, self.model.nv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Collection of open source robot models.

These models are optional and not part of Pink. You can import them by
initializing and updating submodules in a local Pink repository.
"""

import pinocchio as pin

from .jvrc import build_jvrc_from_urdf
from .upkie import build_upkie_from_urdf


class UnknownModel(Exception):

    """
    Exception raised when the model to load is not recognized from its path.
    """


def build_from_urdf(path: str) -> pin.RobotWrapper:
    """
    Build a robot model from its description folder.

    Args:
        path: Path to a robot description folder.

    Note:
        The model to load is recognized from the name of the last folder in the
        path, for example "jvrc_description" or "upkie_description".
    """
    path.rstrip("/")
    if path.endswith("jvrc_description"):
        return build_jvrc_from_urdf(path)
    if path.endswith("upkie_description"):
        return build_upkie_from_urdf(path)
    raise UnknownModel(path)


__all__ = [
    "build_from_urdf",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Upkie wheeled biped bending its knees.
"""

import os
import time

import numpy as np
import pinocchio as pin

import pink
import pink.models

from pink.tasks import BodyTask
from pink import solve_ik

import meshcat_shapes


class ElevatorPose:

    """
    Frame that goes up and down with respect to the world.
    """

    def __init__(self, init: pin.SE3):
        """
        Initialize pose.

        Args:
            init: Initial transform from the wrist frame to the world frame.
        """
        self.init = init

    def at(self, t):
        T = self.init.copy()
        T.translation[2] += 0.1 * np.sin(t)
        return T


if __name__ == "__main__":
    models_dir = os.path.join(os.path.dirname(__file__), "../tests", "models")
    upkie_description = os.path.join(models_dir, "upkie_description")
    robot = pink.models.build_from_urdf(upkie_description)
    viz = pin.visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    robot.setVisualizer(viz, init=False)
    viz.initViewer(open=True)
    viz.loadViewerModel()

    configuration = pink.apply_configuration(robot, robot.q0)
    viz.display(configuration.q)

    base_task = BodyTask("base", position_cost=1.0, orientation_cost=1.0)
    left_contact_task = BodyTask(
        "left_contact", position_cost=1.0, orientation_cost=1.0
    )
    right_contact_task = BodyTask(
        "right_contact", position_cost=1.0, orientation_cost=1.0
    )
    tasks = [base_task, left_contact_task, right_contact_task]

    base_task.set_target(
        configuration.get_transform_body_to_world("base")
    )
    left_contact_task.set_target(
        configuration.get_transform_body_to_world("left_contact")
    )
    right_contact_task.set_target(
        configuration.get_transform_body_to_world("right_contact")
    )

    left_contact_target = ElevatorPose(
        configuration.get_transform_body_to_world("left_contact")
    )
    right_contact_target = ElevatorPose(
        configuration.get_transform_body_to_world("right_contact")
    )

    left_contact_frame = viz.viewer["left_contact_frame"]
    meshcat_shapes.set_frame(left_contact_frame)
    right_contact_frame = viz.viewer["right_contact_frame"]
    meshcat_shapes.set_frame(right_contact_frame)

    dt = 5e-3  # [s]
    for t in np.arange(0.0, 10.0, dt):
        left_contact_task.set_target(left_contact_target.at(t))
        right_contact_task.set_target(right_contact_target.at(t))
        velocity = solve_ik(configuration, tasks, dt)
        q = configuration.integrate(velocity, dt)
        configuration = pink.apply_configuration(robot, q)
        viz.display(q)
        time.sleep(dt)  # TODO(scaron): proper rate



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test fixture for the body task.
"""

import os
import unittest

import numpy as np

from qpsolvers import solve_qp

from pink.tasks import BodyTask
from pink.tasks import TargetNotSet

from .mock_configuration import MockConfiguration


class TestBodyTask(unittest.TestCase):

    """
    Test consistency of the body task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """
        Prepare test fixture.
        """
        models_dir = os.path.join(os.path.dirname(__file__), "models")
        jvrc_description = os.path.join(models_dir, "jvrc_description")
        self.jvrc_description = jvrc_description
        self.mock_configuration = MockConfiguration()

    def test_task_repr(self):
        """
        String representation reports the task gain, costs and target.
        """
        earflap_task = BodyTask(
            "earflap", position_cost=1.0, orientation_cost=0.1
        )
        self.assertTrue("gain=" in repr(earflap_task))
        self.assertTrue("cost=" in repr(earflap_task))
        self.assertTrue("target=" in repr(earflap_task))

    def test_target_not_set(self):
        """
        Raise an exception when the target is not set.
        """
        jetpack_task = BodyTask(
            "jetpack", position_cost=1.0, orientation_cost=0.1
        )
        with self.assertRaises(TargetNotSet):
            jetpack_task.compute_task_dynamics(self.mock_configuration)

    def test_target_set_properly(self):
        """
        Return target properly once it's set.
        """
        jetpack_task = BodyTask(
            "jetpack", position_cost=1.0, orientation_cost=0.1
        )
        T = self.mock_configuration.get_transform_body_to_world("jetpack")
        jetpack_task.set_target(T)
        self.assertIsNotNone(jetpack_task.transform_target_to_world)
        if jetpack_task.transform_target_to_world is None:  # help mypy
            return
        self.assertTrue(
            np.allclose(
                T.homogeneous,
                jetpack_task.transform_target_to_world.homogeneous,
            )
        )

    def test_target_is_a_copy(self):
        """
        Target is saved as a copy, not a reference to the original.
        """
        tail_task = BodyTask("tail", position_cost=1.0, orientation_cost=0.1)
        target = self.mock_configuration.get_transform_body_to_world("tail")
        tail_task.set_target(target)
        y = target.translation[1]
        target.translation[1] += 12.0
        if tail_task.transform_target_to_world is None:  # help mypy
            return
        self.assertAlmostEqual(
            tail_task.transform_target_to_world.translation[1], y
        )
        self.assertNotAlmostEqual(
            tail_task.transform_target_to_world.translation[1],
            target.translation[1],
        )

    def test_zero_error_when_target_at_body(self):
        """
        Error is zero when the target and body are at the same location.
        """
        tail_task = BodyTask("tail", position_cost=1.0, orientation_cost=0.1)
        target = self.mock_configuration.get_transform_body_to_world("tail")
        tail_task.set_target(target)  # error == 0
        J, e = tail_task.compute_task_dynamics(self.mock_configuration)
        self.assertTrue(
            np.allclose(J, self.mock_configuration.get_body_jacobian("tail"))
        )
        self.assertLess(np.linalg.norm(e), 1e-10)

    def test_unit_cost_qp_objective(self):
        """
        A unit cost vector means the QP objective is exactly :math:`(H, c) =
        (J^T J, -e^T J)`.
        """
        shark_task = BodyTask("shark", position_cost=1.0, orientation_cost=0.1)
        target = self.mock_configuration.get_transform_body_to_world("shark")
        shark_task.set_target(target)
        self.mock_configuration.move_body("shark", translate=[0.0, 0.01, 0.0])
        J, e = shark_task.compute_task_dynamics(self.mock_configuration)
        shark_task.set_position_cost(1.0)
        shark_task.set_orientation_cost(1.0)
        shark_task.lm_damping = 0.0
        H, c = shark_task.compute_qp_objective(self.mock_configuration)
        self.assertTrue(np.allclose(J.T @ J, H))
        self.assertTrue(np.allclose(-e.T @ J, c))

    def test_zero_costs_same_as_disabling_lines(self):
        """
        Setting a position or orientation cost to zero yields the same QP
        objective as disabling the corresponding Jacobian and error
        coordinates.
        """
        otter_task = BodyTask("otter", position_cost=1.0, orientation_cost=0.1)
        target = self.mock_configuration.get_transform_body_to_world("otter")
        otter_task.set_target(target)
        self.mock_configuration.move_body("otter", translate=[0.1, 0.02, 0.01])
        J, e = otter_task.compute_task_dynamics(self.mock_configuration)
        qd = np.random.random(J.shape[1:])
        test_cases = {
            "position_only": (1.0, 0.0, slice(0, 3)),
            "orientation_only": (0.0, 1.0, slice(3, 6)),
            "position_0": ([1.0, 0.0, 0.0], 0.0, slice(0, 1)),
            "position_1": ([0.0, 1.0, 0.0], 0.0, slice(1, 2)),
            "position_2": ([0.0, 0.0, 1.0], 0.0, slice(2, 3)),
            "orientation_0": (0.0, [1.0, 0.0, 0.0], slice(3, 4)),
            "orientation_1": (0.0, [0.0, 1.0, 0.0], slice(4, 5)),
            "orientation_2": (0.0, [0.0, 0.0, 1.0], slice(5, 6)),
        }
        for case in test_cases.values():
            position_cost, orientation_cost, indexes = case
            otter_task.set_position_cost(position_cost)
            otter_task.set_orientation_cost(orientation_cost)
            otter_task.lm_damping = 0.0
            H, c = otter_task.compute_qp_objective(self.mock_configuration)
            H_check = J[indexes].T @ J[indexes]
            c_check = -e[indexes].T @ J[indexes]
            cost = qd.T @ H @ qd + c @ qd
            cost_check = qd.T @ H_check @ qd + c_check @ qd
            self.assertAlmostEqual(cost, cost_check)

    def test_lm_damping_has_no_effect_at_target(self):
        """
        Levenberg-Marquardt damping has no effect when the task error is zero.
        """
        leven_task = BodyTask("leven", position_cost=1.0, orientation_cost=0.1)
        target = self.mock_configuration.get_transform_body_to_world("leven")
        leven_task.set_target(target)
        leven_task.lm_damping = 1e-8
        H_1, c_1 = leven_task.compute_qp_objective(self.mock_configuration)
        leven_task.lm_damping = 1e-4
        H_2, c_2 = leven_task.compute_qp_objective(self.mock_configuration)
        self.assertTrue(np.allclose(H_1, H_2))
        self.assertTrue(np.allclose(c_1, c_2))

    def test_lm_damping_has_effect_under_error(self):
        """
        Levenberg-Marquardt damping is indeed a damping: unless the task is
        fulfilled, it reduces velocities.
        """
        berg_task = BodyTask("berg", position_cost=1.0, orientation_cost=0.1)
        target = self.mock_configuration.get_transform_body_to_world("berg")
        berg_task.set_target(target)
        self.mock_configuration.move_body("berg", translate=[0.0, 2.0, 0.0])
        berg_task.lm_damping = 1e-8
        H_1, c_1 = berg_task.compute_qp_objective(self.mock_configuration)
        berg_task.lm_damping = 1e-4
        H_2, c_2 = berg_task.compute_qp_objective(self.mock_configuration)
        qd_1 = solve_qp(H_1, c_1)  # H_1 is p.s.d. from LM damping
        qd_2 = solve_qp(H_2, c_2)  # idem for H_2
        self.assertGreater(np.linalg.norm(qd_2 - qd_1), 1e-6)
